#include <stdio.h>
#include <stddef.h>
#include <string.h>
#ifdef USE_VARARGS
#include <varargs.h>
#else
#include <stdarg.h>
#endif

#ifdef linux
#include <getopt.h>
#endif

#ifdef __svr4__
#include <stdlib.h>
#undef NULL
#define NULL ((void*) 0)
extern char * strdup(const char *);
#endif

char * program="digest";
struct menuitem{
  struct menuitem * parent;
  struct menuitem * next;
  struct menuitem * down;
  int level;
  int defined;
  char * tag;
  char * label;
  char * file;
  char * starting_line;
  char * ending_line;
};

struct menuitem Top = {NULL, NULL, NULL, 0, 0, NULL, NULL, NULL, NULL, NULL};

struct menuitem * root = &Top;
struct menuitem * currentmenu = NULL;

char buffer[1024];
char tag[80], label[132], file[256], starting_line[256], ending_line[256];
int line_number;

FILE * infile, *outfile;

char id[24];

#ifdef USE_VARARGS
void error(va_alist)
  va_dcl
#else
void error(char *format, ...)
#endif
{
  va_list vp;
#ifdef USE_VARARGS
  char *format;
#endif

  fprintf(stderr, "%s: ", program);

#ifdef USE_VARARGS
  va_start(vp);
  format = va_arg (vp, char *);
#else
  va_start(vp, format);
#endif
  vfprintf(stderr, format, vp);
  va_end(va);

  fputc('\n', stderr);

  exit(1);
}

dumpmenu(int indent, struct menuitem * mpnt){
  int i;
  while(mpnt){
    for(i=0; i<indent; i++) fprintf(stderr," ");
    fprintf(stderr,"%d %s:: %s\n", mpnt->level, mpnt->tag, mpnt->label);
    if(mpnt->down) dumpmenu(indent+4, mpnt->down);
    mpnt = mpnt->next;
  };
}


static int lineflag = 0;

void sanitize_and_write(char * string){
  char obuffer[2048];
  char * pnt, *pnt1;
  int flag;

  pnt = string;
  pnt1 = obuffer;
  if(*pnt == '\t') lineflag++;
  flag = 0;
  while(*pnt) {
    if(*pnt == '@') *pnt1++ = '@';
    if(*pnt == '{') *pnt1++ = '@';
    if(*pnt == '}') *pnt1++ = '@';
    if(*pnt == '\t') *pnt = ' ';
    if(*pnt != ' ') flag++;
    *pnt1++ = *pnt++;
  };
  *pnt1++ = 0;
  if(!flag) lineflag = 0;
  if(lineflag && obuffer[0] != ' ') {
    lineflag = 0;
    fprintf(outfile, "\n");
  };
  fprintf(outfile, "%s", obuffer);
}

dumpinfo(struct menuitem * mnode){
  struct menuitem * mprev, *mpnt;
  FILE * infofile;
  int status;
  int writing;
  char ibuffer[2048];
  mprev = NULL;
  mpnt = mnode;


  fprintf(outfile,"@menu\n");
  while(mpnt){
    fprintf(outfile,"* %s::  %s\n", mpnt->tag, mpnt->label);
    mpnt = mpnt->next;
  };
  fprintf(outfile,"@end menu\n");
  mpnt = mnode;
  while(mpnt){
    fprintf(outfile,"@node %s, %s, %s, %s\n", mpnt->tag,
	    (mpnt->next ? mpnt->next->tag: ""),
	    (mprev ? mprev->tag: mpnt->parent->tag),
	    mpnt->parent->tag);
    fprintf(outfile,"@unnumbered %s\n\n", mpnt->label);

    if(strlen(mpnt->file)){
      infofile = fopen(mpnt->file, "r");

      if(infofile){
	if(strlen(mpnt->starting_line) == 0) writing = 1;
	else{
	  writing = 0;
	  status = re_comp(mpnt->starting_line);
	  if(status) {
	    fprintf(stderr,"Invalid regular expression %s\n", mpnt->starting_line);
	    if(strlen(mpnt->ending_line)){
	      writing = 2;
	      status = re_comp(mpnt->ending_line);
	      if(status){
		fprintf(stderr,"Invalid regular expression %s\n",
			mpnt->ending_line);
		writing = 1;
	      } else
		writing = 1;
	      };
	  };
	};
	while(1==1){
	  fgets(ibuffer, sizeof(ibuffer), infofile);
	  if(feof(infofile)) break;
	  if(writing == 2) {
	    if(re_exec(ibuffer) == 1) {
	      break;
	    };
	  };
	  if(writing) sanitize_and_write(ibuffer);
	  if(!writing) {
	    if(re_exec(ibuffer) == 1) {
	      if(strlen(mpnt->ending_line) == 0) writing = 1;
	      else {
		writing = 2;
		status = re_comp(mpnt->ending_line);
		if(status){
		  fprintf(stderr,"Invalid regular expression %s\n",
			  mpnt->ending_line);
		  writing = 1;
		};
	      };
	    };
	  };
	};
	fclose(infofile);
      };
    };
    if(mpnt->down) dumpinfo(mpnt->down);
    mprev = mpnt;
    mpnt = mpnt->next;
  }
}

dumpinfo_header(){
  fprintf(outfile,"\\input texinfo    @c -*-texinfo-*-\n");
  fprintf(outfile,"@comment %%**start of header\n");
  fprintf(outfile,"@setfilename linux.info\n");
  fprintf(outfile,"@settitle Linux @value{edition}\n");
  fprintf(outfile,"@syncodeindex vr fn\n");
  fprintf(outfile,"@c footnotestyle separate\n");
  fprintf(outfile,"@c paragraphindent 2\n");
  fprintf(outfile,"@smallbook\n");
  fprintf(outfile,"@comment %%**end of header\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@ignore\n");
  fprintf(outfile,"@ifinfo\n");
  fprintf(outfile,"@format\n");
  fprintf(outfile,"START-INFO-DIR-ENTRY\n");
  fprintf(outfile,"* Linux: (linux.info).           Linux FAQ.\n");
  fprintf(outfile,"END-INFO-DIR-ENTRY\n");
  fprintf(outfile,"@end format\n");
  fprintf(outfile,"@end ifinfo\n");
  fprintf(outfile,"@end ignore\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@c Set smallbook if printing in smallbook format so the example of the\n");
  fprintf(outfile,"@c smallbook font is actually written using smallbook; in bigbook, a kludge\n");
  fprintf(outfile,"@c is used for TeX output.\n");
  fprintf(outfile,"@c set smallbook\n");
  fprintf(outfile,"@clear smallbook\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@set edition 0.85\n");
  fprintf(outfile,"@set update-date 13 Oct 1993\n");
  fprintf(outfile,"@set update-month October 1993\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@c Experiment with smaller amounts of whitespace between chapters\n");
  fprintf(outfile,"@c and sections.\n");
  fprintf(outfile,"@tex\n");
  fprintf(outfile,"\\global\\chapheadingskip = 15pt plus 4pt minus 2pt \n");
  fprintf(outfile,"\\global\\secheadingskip = 12pt plus 3pt minus 2pt\n");
  fprintf(outfile,"\\global\\subsecheadingskip = 9pt plus 2pt minus 2pt\n");
  fprintf(outfile,"@end tex\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@c Experiment with smaller amounts of whitespace between paragraphs in\n");
  fprintf(outfile,"@c the 8.5 by 11 inch format.\n");
  fprintf(outfile,"@ifclear smallbook\n");
  fprintf(outfile,"@tex\n");
  fprintf(outfile,"\\global\\parskip 6pt plus 1pt\n");
  fprintf(outfile,"@end tex\n");
  fprintf(outfile,"@end ifclear\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@finalout\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@ifinfo\n");
  fprintf(outfile,"This provides a hypertext interface to the Linux FAQ.\n");
  fprintf(outfile,"@setchapternewpage odd\n");
  fprintf(outfile,"@titlepage\n");
  fprintf(outfile,"@c use the new format for titles\n");
  fprintf(outfile,"@title LinuxFAQ\n");
  fprintf(outfile,"@subtitle The Linux Documentation Project\n");
  fprintf(outfile,"@subtitle Edition @value{edition}, for the Linux FAQ.\n");
  fprintf(outfile,"@subtitle @value{update-month}\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"@author by a cast of thousands.\n");
  fprintf(outfile,"@end titlepage\n");
#if 0
  fprintf(outfile,"@ifinfo\n");
  fprintf(outfile,"@node Top, %s, (dir), (dir)\n", Top.down->tag);
  fprintf(outfile,"@top Linux FAQ\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"	The hyperFAQ is a quick and easy way for people to browse the\n");
  fprintf(outfile,"Linux FAQ.  It contains the answers to all of the frequently asked\n");
  fprintf(outfile,"questions.  @refill\n");
  fprintf(outfile,"\n");
  fprintf(outfile,"This is Edition @value{edition} of the linux hyper-FAQ, \n");
  fprintf(outfile,"@w{@value{update-date},} for Texinfo Version 2.\n");
  fprintf(outfile,"@end ifinfo\n");
#endif
}

dumpinfo_trailer(){
  fprintf(outfile,"@summarycontents\n");
  fprintf(outfile,"@contents\n");
  fprintf(outfile,"@bye\n");
}

void findmenu(char * menuid){
  char * pnt, *pnt1;
  struct menuitem * mpnt, *mpnt1;
  int level;
  int i;

  currentmenu = root;
  strcpy(id, menuid);
  pnt = id;
  while(strlen(pnt)){
    pnt1 = strchr(pnt,'.');
    if(pnt1) *pnt1 = 0;
    if(pnt[0] > '9' || pnt[0] < '0') {
      level = 0;
      for(i=0; pnt[i] && i<4; i++) level = (level << 8) + pnt[i];
    } else
      sscanf(pnt, "%d", &level);
    mpnt = currentmenu->down;
    while(mpnt){
      if(mpnt->level == level) break;
      mpnt = mpnt->next;
    };
    if(!mpnt){
      mpnt = (struct menuitem *) malloc(sizeof(struct menuitem));
      mpnt->down = NULL;
      mpnt->parent = currentmenu;
      mpnt->next = NULL;
      if(!currentmenu->down){
	currentmenu->down = mpnt;
      } else {
	mpnt1 = currentmenu->down;
	while(mpnt1->next) mpnt1=mpnt1->next;
	mpnt1->next = mpnt;
      };
      mpnt->level = level;
      mpnt->defined = 0;
      mpnt->tag = strdup("");
      mpnt->file = strdup("");
      mpnt->label = strdup("");
      mpnt->starting_line = strdup("");
      mpnt->ending_line = strdup("");;
    };
    currentmenu = mpnt;
    if(!pnt1 && mpnt->defined) 
      fprintf(stderr,"Warning, menu %s redefined\n", menuid);

    mpnt->defined = 1;
    if(!pnt1) break;
    pnt = pnt1+1;
  };
}

int main(int argc, char * argv[]){
  char * bpnt;
  char c;

  if(argc<2) exit(1);
  outfile = stdout;
  Top.tag = strdup("Top");
  Top.file = strdup("");
  Top.label=strdup("");
  Top.starting_line=strdup("");
  Top.ending_line=strdup("");

  while ((c = getopt(argc, argv, "o:")) != EOF)
    {
      switch (c)
	{
	case 'o':
	  outfile = fopen(optarg,"w");
	  break;
	default:
	  error("invalid option '-%c'", c);
	  exit(1);
	}
    }

  while(optind < argc){
    fprintf(stderr,"Processing file %s\n", argv[optind]);
    infile = fopen(argv[optind++],"r");
    
    while(1==1){
      fgets(buffer, sizeof(buffer), infile);
      line_number++;
      if(feof(infile)) break;
      if(buffer[strlen(buffer)-1] == '\n') buffer[strlen(buffer)-1] = 0;
      
      bpnt = buffer;
      while(*bpnt == ' ' || *bpnt == '\t') bpnt++;
      if(!*bpnt) continue;
      if(*bpnt == '#') continue;  /* Comments */
      
      if(strncmp(bpnt,"menupos=", 8) == 0){
	findmenu(bpnt+8);
	continue;
      };
      
      if(currentmenu){
	if(strncmp(bpnt,"tag=", 4) == 0){
	  if(currentmenu->tag) free(currentmenu->tag);
	  currentmenu->tag = strdup(bpnt+4);
	  continue;
	};
	
	if(strncmp(bpnt,"label=", 6) == 0){
	  if(currentmenu->label) free(currentmenu->label);
	  currentmenu->label = strdup(bpnt+6);
	  continue;
	};
	
	if(strncmp(bpnt,"file=", 5) == 0){
	  if(currentmenu->file) free(currentmenu->file);
	  currentmenu->file = strdup(bpnt+5);
	  continue;
	};
	
	if(strncmp(bpnt,"starting_line=", 14) == 0){
	  if(currentmenu->starting_line) free(currentmenu->starting_line);
	  currentmenu->starting_line = strdup(bpnt+14);
	  continue;
	};
	
	if(strncmp(bpnt,"ending_line=", 12) == 0){
	  if(currentmenu->ending_line) free(currentmenu->ending_line);
	  currentmenu->ending_line = strdup(bpnt+12);
	  continue;
	};
      };      
      fprintf(stderr,"Syntax error at line %d\n", line_number);
    };
  };
  /*  dumpmenu(0, &Top); */
  dumpinfo_header();
  dumpinfo(&Top);
  dumpinfo_trailer();
  return 0;
}
